///////////////////////////////////////////////////////////
//     TKnob Delphi Visual Component Library
//     by I MD. CIPTAYASA(IMDC)
//     ******************************
//     Description : Manual_Radio_Tuner_Panel-like VCL
//     Version     : 1.0.0
//     Base Class  : TCustomControl
//     Programmer  : IMDC
//     Date        : Jan'2001
//     Compiler    : Delphi 3 (it'd be straightforward task porting to Delphi 5)
//
//     ****************************************************
//     Feel free to modify to fit your needs
//////////////////////////////////////////////////////////


unit Knob;

interface

uses
  Windows, Messages, SysUtils, Classes, Graphics, Controls, Forms, Dialogs,
  DsgnIntf;

type
  TRollingEvent = procedure(Position : integer) of object;
  TKnob = class(TCustomControl)
  private
    FFrame      : Boolean;
    FFrameColor : TColor;
    FBackColor  : TColor;
    FCaption    : TCaption;
    FAngle      : real;
    FPosition   : integer;
    FMinValue   : integer;
    FMaxValue   : integer;
    FFont       : TFont;
    FRegion     : HRGN;
    FPosRect    : TRect;
    FDrawingArea: TRect;
    FOldPos     : TPoint;
    FDiffX      : integer;
    FDiffY      : integer;
    FOnRolling  : TRollingEvent;
    FKeySteps   : Cardinal;
    FTicks      : Cardinal;
    FActiveColor: TColor;

    procedure PlaceLabel(ABitmap : TBitmap;X,Y : integer; S : string);
    procedure GetDrawingArea(var Area : TRect);
    procedure DrawMarkPosition(ACanvas:TCanvas;Position:integer;Erase:Boolean);
    procedure SetActiveColor(Value : TColor);
    procedure SetFrameColor(Value:TColor);
    procedure SetMinValue(Value:integer);
    procedure SetMaxValue(Value:integer);
    procedure SetPosition(Value:integer);
    procedure SetFrame(Value : Boolean);
    procedure SetBackColor(Value : TColor);
    procedure SetKeySteps(Value : cardinal);
    procedure SetCaption(Value : TCaption);
    procedure SetTicks(Value : Cardinal);
    procedure DrawFace(ABitmap:TBitmap);

  protected
    procedure Click;override;
    procedure Paint;override;
    procedure WMSize(var Message:TMessage);message WM_SIZE;
    procedure MouseDown(Button: TMouseButton; Shift: TShiftState; X, Y: Integer);override;
    procedure MouseMove(Shift: TShiftState; X, Y: Integer);override;
    procedure MouseUp(Button: TMouseButton; Shift: TShiftState; X, Y: Integer);override;
    procedure KeyDown(var Key: Word; Shift: TShiftState);override;
    procedure KeyUp(var Key: Word; Shift: TShiftState);override;
    procedure CMEnter(var Message:TMessage);message CM_ENTER;
    procedure CMExit(var Message:TMessage);message CM_EXIT;
    procedure CMFontChanged(var Message:TMessage);message CM_FONTCHANGED;
    procedure WMDestroy(var Message:TMessage);message WM_DESTROY;
  public
    constructor Create(AOwner:TComponent);override;
  published
    property ActiveColor : TColor read FActiveColor write SetActiveColor default clLime;
    property FrameColor:TColor read FFrameColor write SetFrameColor default clAqua;
    property BackGroundColor:TColor read FBackColor write SetBackColor default clBtnFace;
    property MinValue:integer read FMinValue write SetMinValue;
    property Caption:TCaption read FCaption write SetCaption;
    property MaxValue:integer read FMaxValue write SetMaxValue;
    property Position:integer read FPosition write SetPosition;
    property Steps: cardinal read FKeySteps write SetKeySteps default 2;
    property Enabled;
    property Framed : Boolean read FFrame write SetFrame default false;
    property TicksFreq:Cardinal read FTicks write SetTicks;
    property Font;
    property ParentColor;
    property ParentFont;
    property ParentShowHint;
    property ShowHint;
    property Visible;
    property OnClick;
    property OnKeyDown;
    property OnKeyUp;
    property OnMouseDown;
    property OnMouseUp;
    property OnRolling:TRollingEvent read FOnRolling write FOnRolling;
  end;

  TChipEditor = class(TComponentEditor)
    procedure Edit;override;
  end;

procedure Register;

implementation

procedure TChipEditor.Edit;
begin
  ShowMessage('TKnob - Delphi VCL'+#13#10+'Free Delphi VCL by IMDC');
end;

constructor TKnob.Create(AOwner:TComponent);
begin
  inherited Create(AOwner);
  FFrame      := false;
  FTicks      := 10;
  FBackColor  := clBtnFace;
  FFrameColor := clAqua;
  FActiveColor := clLime;
  FMinValue   := 0;
  FMaxValue   := 100;
  Width       := 150;
  Height      := 150;
  FPosition   := 0;
  TabStop     := true;
  FKeySteps   := 5;
  FFont       := Font;
  FFont.Color := clRed;
  ParentColor := true;
  ParentFont  := false;
end;

procedure TKnob.GetDrawingArea(var Area : TRect);
var
 R : TRect;
 W,H : integer;
begin
 R := ClientRect;
 W := R.Right - R.Left;
 H := R.Bottom - R.Top;

 // Make Width = Height

 if W > H then
    R.Right := H
 else
    R.Bottom := W;
 CopyRect(Area,R);
end;

function MapToAngle(sMin,sMax,dMin,dMax,Value : integer):integer;
var
 Temp : real;
begin
 Temp := dMin +((dMax - dMin)/(sMax - sMin))*(Value-sMin);
 Result := Round(Temp);
end;

procedure TKnob.DrawMarkPosition(ACanvas:TCanvas;Position:integer;Erase:Boolean);
var
 R : TRect;
 Diff,CX,CY,Radius : integer;
 Xi,Yi,Xo,Yo : integer;
 Teta : real;
begin
  GetDrawingArea(R);
  InflateRect(R,-40,-40);
  Radius := (R.Right-R.Left) div 2;
  Diff := Radius div 8;
  CX:=R.Left + Radius;
  CY:=R.Top +Radius;
  InflateRect(R,-Diff,-Diff);
  Radius := (R.Right-R.Left) div 2;
  FAngle := MapToAngle(FMinValue,FMaxValue,-135,135,Position);
  Teta := Pi/180;
  Xi:=Round((Radius -(Diff*2))*Sin(FAngle*Teta));
  Yi:=Round((Radius -(Diff*2))*Cos(FAngle*Teta));
  Xo:=Round((Radius)*Sin(Fangle*Teta));
  Yo:=Round((Radius)*Cos(FAngle*Teta));

  with ACanvas do
   begin
    if Erase then
     Pen.Color := clBtnFace
    else
    if Focused then
     Pen.Color := FActiveColor
    else
     Pen.Color := clBlack;
     Pen.Width := 3;
     MoveTo(CX+Xo,CY-Yo);
     LineTo(CX+Xi,CY-Yi);
   end;
end;

procedure TKnob.PlaceLabel(ABitmap : TBitmap;X,Y : integer;S : string);
var
 W,H : integer;
begin
 W := ABitmap.Canvas.TextWidth(S);
 H := ABitmap.Canvas.TextHeight('H');
 with ABitmap.Canvas do
  begin
    Font := FFont;
    Brush.Style := bsClear;
    TextOut(X - (W div 2),Y-(H div 2),S);
  end;
end;

procedure TKnob.DrawFace(ABitmap:TBitmap);
var
 i                : integer;
 Off,Radius       : integer;
 CX,CY,Xo,Yo,Xi,Yi: integer;
 RFrame,ROuter    : TRect;
 Teta             : real;
 Num              : integer;
begin
  GetDrawingArea(RFrame);
  GetDrawingArea(ROuter);
  InflateRect(ROuter,-20,-20);
  Radius:=(ROuter.Right-ROuter.Left) div 2;
  Off := Radius div 8;
  CX:=ROuter.Left+Radius;
  CY:=ROuter.Top+Radius;
  FPosRect:=Rect(CX+(Radius div 2)-10,CY+(Radius div 2)-10,CX+(Radius div 2),CY+(Radius div 2));
  InflateRect(ROuter,-8,-8);
  with ABitmap.Canvas do
   begin
    Brush.Color := FBackColor;
    Pen.Color:=FFrameColor;
    Pen.Width:=1;
    if FFrame = true then
     begin
       Rectangle(RFrame.Left,RFrame.Top,RFrame.Right,RFrame.Bottom);
     end;
    Brush.Style := bsClear;
    Arc(ROuter.Left,ROuter.Top,ROuter.Right,ROuter.Bottom,
        ROuter.Right,ROuter.Bottom,ROuter.Left,ROuter.Bottom);
    InflateRect(ROuter,-Off,-Off);
    Brush.Style := bsSolid;
    Pen.Color:=clBtnFace;
    Brush.Color:=clBtnFace;
    Ellipse(ROuter.Left,ROuter.Top,ROuter.Right,ROuter.Bottom);
    Pen.Width:=3;
    Pen.Color:=clBtnHighLight;
    Arc(ROuter.Left,ROuter.Top,ROuter.Right,ROuter.Bottom,
        ROuter.Right,ROuter.Top,ROuter.Left,ROuter.Bottom);
    Pen.Color:=clBtnShadow;
    Arc(ROuter.Left,ROuter.Top,ROuter.Right,ROuter.Bottom,
        ROuter.Left,ROuter.Bottom,ROuter.Right,ROuter.Top);
    Pen.Color:=FFrameColor;
    Pen.Width:=1;
    Teta:= pi/180;
    for i:=-135 to 135 do
     begin
       Xo:=Round((Radius-5)*Sin(i*Teta));
       Yo:=Round((Radius-5)*Cos(i*Teta));
       Xi:=Round((Radius-10)*Sin(i*Teta));
       Yi:=Round((Radius-10)*Cos(i*Teta));
       if (i mod FTicks)=0 then
        begin
         MoveTo(CX+Xo,CY-Yo);
         LineTo(CX+Xi,CY-Yi);
        end;
     end;
       Xo:=Round(Radius*Sin(-135*Teta));
       Yo:=Round(Radius*Cos(-135*Teta));
       Xi:=Round((Radius-10)*Sin(-135*Teta));
       Yi:=Round((Radius-10)*Cos(-135*Teta));
       MoveTo(CX+Xo,CY-Yo);
       LineTo(CX+Xi,CY-Yi);
       PlaceLabel(ABitmap,CX+Xo-5,CY-Yo+5,IntToStr(FMinValue));

       Xo:=Round(Radius*Sin(0*Teta));
       Yo:=Round(Radius*Cos(0*Teta));
       Xi:=Round((Radius-10)*Sin(0*Teta));
       Yi:=Round((Radius-10)*Cos(0*Teta));
       MoveTo(CX+Xo,CY-Yo);
       LineTo(CX+Xi,CY-Yi);
       Num := (FMaxValue + FMinValue) div 2;
       PlaceLabel(ABitmap,CX,CY-Yo-5,IntToStr(Num));
       //Draw Caption
       PlaceLabel(ABitmap,CX,CY+Yo+5,FCaption);

       Xo:=Round(Radius*Sin(135*Teta));
       Yo:=Round(Radius*Cos(135*Teta));
       Xi:=Round((Radius-10)*Sin(135*Teta));
       Yi:=Round((Radius-10)*Cos(135*Teta));
       MoveTo(CX+Xo,CY-Yo);
       LineTo(CX+Xi,CY-Yi);
       PlaceLabel(ABitmap,CX+Xo+5, CY-Yo+5,IntToStr(FMaxValue));
   end;
   DrawMarkPosition(ABitmap.Canvas,FPosition,false);
end;

procedure TKnob.Paint;
var
 OffScreen:TBitmap;
begin
// Create OffScreen buffer to enhance performance
  try
   OffScreen:=TBitmap.Create;
   OffScreen.Width:=Width;
   OffScreen.Height:=Height;
   OffScreen.Canvas.Handle:=Canvas.Handle;
   DrawFace(OffScreen);

   // Copy back to screen >><< BitBlt is faster than canvas Draw method

   BitBlt(Canvas.Handle,0,0,Width,Height,OffScreen.Canvas.Handle,0,0,SRCCOPY);
   finally
    OffScreen.Free;
  end;
  inherited Paint;
end;

procedure TrapEllipticMovement(var OldX,OldY :integer;CurrentX,CurrentY : integer;
                               var DiffX,DiffY : integer);
begin
  DiffX := CurrentX - OldX;
  DiffY := CurrentY - OldY;
end;

procedure LimitPosition(Min,Max : integer;var Value : integer);
begin
  if Value <= Min then
   Value := Min;
  if Value >= Max then
   Value := Max;
end;

procedure TKnob.MouseDown(Button: TMouseButton; Shift: TShiftState; X, Y: Integer);
var
 R:TRect;
begin
  inherited MouseDown(mbLeft,[ssLeft],X,Y);
  FOldPos := Point(X,Y);
  GetDrawingArea(R);
  InflateRect(R,-28,-28);
  FRegion := CreateEllipticRgn(R.Left,R.Top,R.Right,R.Bottom);
  SelectClipRgn(Canvas.Handle,FRegion);
end;

procedure TKnob.MouseMove(Shift: TShiftState; X, Y: Integer);
var
 Center : TPoint;
begin
  inherited MouseMove(Shift,X,Y);
  GetDrawingArea(FDrawingArea);
  Center.X := (FDrawingArea.Right - FDrawingArea.Left) div 2;
  Center.Y := (FDrawingArea.Bottom - FDrawingArea.Top) div 2;
  if (ssLeft in Shift) and (PtInRegion(FRegion,X,Y)) then
   begin
    TrapEllipticMovement(FOldPos.X,FOldPos.Y,X,Y,FDiffX,FDiffY);
    if (X < Center.X) then
      begin
       DrawMarkPosition(Canvas,FPosition,true);
       FPosition := FPosition - FKeySteps*FDiffY;
       LimitPosition(FMinValue,FMaxValue,FPosition);
       DrawMarkPosition(Canvas,FPosition,false);
      end;
    if (X > Center.X) then
      begin
       DrawMarkPosition(Canvas,FPosition,true);
       FPosition := FPosition + FKeySteps*FDiffY;
       LimitPosition(FMinValue,FMaxValue,FPosition);
       DrawMarkPosition(Canvas,FPosition,false);
      end;
   end;
   FOldPos.X := X;
   FOldPos.Y := Y;
  if Assigned(FOnRolling) then FOnRolling(FPosition);

end;

procedure TKnob.MouseUp(Button: TMouseButton; Shift: TShiftState; X, Y: Integer);
begin
  inherited MouseUp(mbLeft,[ssLeft],X,Y);
end;

procedure TKnob.KeyDown(var Key: Word; Shift: TShiftState);
begin
  inherited KeyDown(Key,Shift);
  case Key of
  VK_PRIOR :
   begin
     DrawMarkPosition(Canvas,FPosition,true);
     Inc(FPosition,FKeySteps);
     LimitPosition(FMinValue,FMaxValue,FPosition);
     DrawMarkPosition(Canvas,FPosition,false);
   end;
   VK_NEXT :
    begin
     DrawMarkPosition(Canvas,FPosition,true);
     Dec(FPosition,FKeySteps);
     LimitPosition(FMinValue,FMaxValue,FPosition);
     DrawMarkPosition(Canvas,FPosition,false);
    end;
   VK_HOME :
    begin
      DrawMarkPosition(Canvas,FPosition,true);
      FPosition := FMinValue;
      DrawMarkPosition(Canvas,FPosition,false);
    end;
   VK_END :
    begin
      DrawMarkPosition(Canvas,FPosition,true);
      FPosition := FMaxValue;
      LimitPosition(FMinValue,FMaxValue,FPosition);
      DrawMarkPosition(Canvas,FPosition,false);
    end;
  end;
  if Assigned(FOnRolling) then FOnRolling(FPosition);
end;

procedure TKnob.KeyUp(var Key: Word; Shift: TShiftState);
begin
  inherited KeyUp(Key,Shift);
end;

procedure TKnob.CMEnter(var Message:TMessage);
begin
  DrawMarkPosition(Canvas,FPosition,false);
end;

procedure TKnob.CMExit(var Message:TMessage);
begin
  DrawMarkPosition(Canvas,FPosition,false);
  DeleteObject(FRegion);
end;

procedure TKnob.CMFontChanged(var Message:TMessage);
begin
   Refresh;
end;

procedure TKnob.WMDestroy(var Message:TMessage);
begin
  inherited;
  DeleteObject(FRegion);
end;

procedure TKnob.Click;
begin
  inherited Click;
  SetFocus;
end;

procedure TKnob.SetActiveColor(Value : TColor);
begin
  if Value <> FActiveColor then
   begin
    FActiveColor := Value;
    Refresh;
   end;
end;

procedure TKnob.SetFrame(Value : Boolean);
begin
   if Value <> FFrame then
    begin
      FFrame := Value;
      Refresh;
    end;
end;

procedure TKnob.SetCaption(Value : TCaption);
begin
   if Value <> FCaption then
    begin
      FCaption := Value;
      Refresh;
    end;
end;

procedure TKnob.SetFrameColor(Value:TColor);
begin
  if Value<>FFrameColor then
   begin
    FFrameColor := Value;
    Refresh;
   end;
end;

procedure TKnob.SetMinValue(Value:integer);
begin
  if Value<>FMinValue then
   begin
    FMinValue:=Value;
    Refresh;
   end;
end;

procedure TKnob.SetMaxValue(Value:integer);
begin
  if Value<>FMaxValue then
   begin
    FMaxValue:=Value;
    Refresh;
   end;
end;

procedure TKnob.SetBackColor(Value : TColor);
begin
  if Value <> FBackColor then
   begin
    FBackColor := Value;
    Refresh;
   end;
end;

procedure TKnob.SetPosition(Value:integer);
begin
  if (Value >= FMinValue) and (Value <=FMaxValue) then
   begin
    DrawMarkPosition(Canvas,FPosition,true);
    FPosition:=Value;
    DrawMarkPosition(Canvas,FPosition,false);
   end;
end;

procedure TKnob.SetKeySteps(Value : cardinal);
begin
  if Value <> FKeySteps then
   begin
    FKeySteps := Value;
   end;
end;

procedure TKnob.WMSize(var Message:TMessage);
begin
  if Width < 120 then
   Width := 120;
  if Height < 120 then
   Height := 120;
end;

procedure TKnob.SetTicks(Value : Cardinal);
begin
  if Value <> FTicks then
   begin
    FTicks :=  Value;
    Refresh;
   end;
end;

procedure Register;
begin
  RegisterComponents('Chip2000', [TKnob]);
  RegisterComponentEditor(TKnob,TChipEditor);
end;

end.
